"""
Abstract base classes for the ukrdc_stats package
"""

from abc import ABC, abstractmethod
from typing import Optional, List

import pandas as pd
import datetime as dt
from sqlalchemy.orm import Session
from sqlalchemy import select, func
from ukrdc_sqla.ukrdc import PatientRecord, Patient

from ukrdc_stats.models.base import JSONModel
from ukrdc_stats.models.generic_2d import BaseTable
from ukrdc_stats.exceptions import NoCohortError

from ukrdc_sqla.ukrdc import PatientNumber
from ukrdc_stats.utils import GENDER_GROUP_MAP, map_codes


class AbstractFacilityStatsCalculator(ABC):
    """
    Abstract base class for facility stats calculators.

    We only enforce a couple of minor requirements:
    - The class must have a constructor that takes a database session and a facility code
    - The class must have an `extract_patient_cohort` method that assigns a pandas dataframe to the `_patient_cohort` attribute
    - The class must have a `calculate_stats` method that returns calculated stats as a pydantic model
    """

    def __init__(
        self,
        session: Session,
        facility: str,
        date: Optional[dt.datetime] = None,
    ):
        # Set up the database session
        self.session: Session = session

        # Store the facility code
        self.facility: str = facility

        # Create a pandas dataframe to store the results
        self._patient_cohort: Optional[pd.DataFrame] = None

        self.date = date or dt.datetime.now()

    def produce_report(
        self,
        output_columns: List[str],
        input_filters: list[str] = None,
        include_ni: bool = False,
    ) -> BaseTable:
        """
        Produce report containing the patients from a cohort displayed in the
        as aggregated stats. As UI users can't query patients on the pid they
        should probably be returned as a list of mrns.
        """
        if self._patient_cohort is None:
            self.extract_patient_cohort()

        if self._patient_cohort is None:
            raise NoCohortError

        if "ukrdcid" not in output_columns:
            output_columns.append("ukrdcid")

        if input_filters:
            dataframe_filter = "(" + ")&(".join(input_filters) + ")"
            patient_record_filtered = self._patient_cohort.query(dataframe_filter)
        else:
            patient_record_filtered = self._patient_cohort

        population = len(patient_record_filtered.ukrdcid.drop_duplicates())

        report = (
            patient_record_filtered[output_columns]
            .drop_duplicates()
            .reset_index(drop=True)
        )

        if include_ni:
            patient_numbers = pd.DataFrame(
                self.session.execute(
                    select(PatientRecord.ukrdcid, PatientNumber.patientid)
                    .join(PatientRecord, PatientNumber.pid == PatientRecord.pid)
                    .where(
                        PatientNumber.organization == "NHS",
                        PatientRecord.ukrdcid.in_(report.ukrdcid.drop_duplicates()),
                    )
                ),
            ).rename(columns={"patientid": "nhsno"})

            report = pd.merge(report, patient_numbers, on="ukrdcid", how="left")
            report["nhsno"] = report["nhsno"].fillna("Unknown")
            report = report.drop_duplicates()

        return population, BaseTable(
            headers=report.columns.tolist(),
            rows=[row.tolist() for _, row in report.iterrows()],
        )

    # Constrain extract patient cohort with
    def extract_patient_cohort(self, **kwargs):
        self._extract_patient_cohort(**kwargs)
        self._validate_patient_cohort(self._patient_cohort)

    @abstractmethod
    def _extract_patient_cohort(self) -> None:
        """
        Extract the patient cohort from the database, and assign it to self._patient_cohort
        """

    def _validate_patient_cohort(self, dataframe: pd.DataFrame):
        """
        There are certain constraint
        """
        if self._patient_cohort is None:
            raise ValueError(
                "Patient cohort extractor has failed to produce the _patient_cohort dataframe"
            )

        required_columns = {"pid", "ukrdcid", "sendingfacility"}
        missing_columns = required_columns - set(dataframe.columns)
        if missing_columns:
            raise ValueError(f"Missing columns in patient cohort: {missing_columns}")

    # All calculators should contain a method to extract aggregated stats as a
    # pydantic object for api of the UI to use
    @abstractmethod
    def extract_stats(self) -> JSONModel:
        """
        Extract all stats from the patient cohort and return them in a Pydantic object

        Returns:
            JSONModel: Pydantic object containing all related stats
        """

    def _query_demographics(self, patient_list):
        """Generalized query to return the demographic information

        Args:
            patient_list (_type_): _description_

        Returns:
            _type_: _description_
        """

        query_demog = (
            select(
                PatientRecord.pid,
                Patient.gender,
                Patient.ethnicgroupcode,
                func.date(Patient.birthtime).label("dob"),
                func.date_part("year", func.age(self.date, Patient.birthtime)).label(
                    "age"
                ),
                func.coalesce(Patient.deathtime > self.date, True).label("is_alive"),
            )
            .join(PatientRecord, Patient.pid == PatientRecord.pid)
            .where(
                PatientRecord.pid.in_(patient_list),
            )
        )

        return pd.DataFrame(self.session.execute(query_demog)).drop_duplicates()

    def _relabel_demographics(self, demographics_raw):
        # Map age to fixed bins
        demographics_raw["age"] = demographics_raw["age"].astype(int)
        bins = [0, 18, 35, 55, 75, 150]
        labels = ["<18", "18-34", "35-54", "55-74", ">=75"]
        for i in range(len(labels)):
            demographics_raw.loc[
                (bins[i + 1] > demographics_raw["age"])
                & (bins[i] <= demographics_raw["age"]),
                "agerange",
            ] = labels[i]

        demographics_raw.loc[~demographics_raw["is_alive"], "agerange"] = "Deceased"
        demographics_raw.loc[demographics_raw["agerange"] == "Deceased", "age"] = None

        # Map gender to stats groups
        demographics_raw["gender"] = (
            demographics_raw["gender"]
            .map(GENDER_GROUP_MAP, None)
            .fillna("Unknown/Uncoded")
        )

        # Map ethnicity to stats groups
        ethnic_group_map = map_codes(
            "NHS_DATA_DICTIONARY", "URTS_ETHNIC_GROUPING", self.session
        )
        demographics_raw["ethnicity"] = (
            demographics_raw["ethnicgroupcode"]
            .map(ethnic_group_map, None)
            .fillna("Unknown/Uncoded")
        )

        return demographics_raw

    def append_demographics(self):
        if self._patient_cohort is None:
            raise ValueError(
                "Please extract cohort before appending demographic information"
            )

        # get demographic attributes from database

        # break up large queries into chunks to avoid PostgreSQL stack overflow
        batch_size = 1000
        demographics = []
        for i in range(0, len(self._patient_cohort), batch_size):
            batch = self._patient_cohort.iloc[i : i + batch_size]
            demographics.append(self._query_demographics(batch.pid))

        demographics = self._relabel_demographics(pd.concat(demographics))

        # merge back into patient cohort and remove duplicated columns
        self._patient_cohort = self._patient_cohort.merge(
            demographics, how="left", on="pid", suffixes=("", "_extra")
        )
        self._patient_cohort = self._patient_cohort.loc[
            :, ~self._patient_cohort.columns.str.endswith("_extra")
        ]

    # For future implementation postcode lookup against external api
    def append_deprivation_index(self):
        pass

    # For future implementation to cache cohorts to improve retrieval speed
    def store(self):
        pass

    def retrieve(self):
        pass



"""
Base Pydantic class with JSON aliasing, used in the UKRDC API
"""

from pydantic import BaseModel


def _to_camel(snake_str: str) -> str:
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


class JSONModel(BaseModel):
    class Config:
        alias_generator = _to_camel
        allow_population_by_field_name = True



"""
Pydantic classes for statistics which generate network graphs.
"""

from typing import List, Optional
from pydantic import Field

from ukrdc_stats.models.base import JSONModel


class NetworkMetaData(JSONModel):
    """Generic class to hold the metadata for data/stats with a network like structure.
    this is used to generate api calls for sankey diagrams
    """

    title: str = Field(..., description="Title of plot or statistic", max_length=40)
    summary: str = Field(..., description="Summary of plot", max_length=100)
    description: str = Field(
        ..., description="In depth description of how statistics are calculated"
    )
    total_population: Optional[int] = Field(
        None, description="Size of patient population for calculated statistic"
    )


class Nodes(JSONModel):
    """Contains labels/names accosiated with each node in the network"""

    node_labels: Optional[List[str]] = None


class Connections(JSONModel):
    """Contains data to describe the links in a sankey diagram (network) these links
    (or edges) are described by three numbers where the link is coming from where it
    is going and the weight"""

    source: List[str] = Field(..., description="index of connection source")
    target: List[str] = Field(..., description="index of connection target")
    value: List[str] = Field(..., description="weight of connection")


class LabelledNetwork(JSONModel):
    metadata: NetworkMetaData = Field(
        ..., description="Meta data containing title description etc"
    )
    node: Nodes = Field(..., description="Information on the nodes of the network")
    link: Connections = Field(..., description="Information on the network links")



"""
Patient cohort dialysis stats calculator
"""

import warnings

import datetime as dt


from typing import Optional, Tuple, List, Dict

import pandas as pd
from sqlalchemy import and_, func, or_, select, cast, case, exists
from sqlalchemy.orm import Session, aliased
from sqlalchemy.types import Float
from ukrdc_sqla.ukrdc import (
    DialysisSession,
    Patient,
    PatientRecord,
    Treatment,
    ModalityCodes,
)

from ukrdc_stats.calculators.abc import AbstractFacilityStatsCalculator
from ukrdc_stats.utils import _get_satellite_list
from ukrdc_stats.exceptions import NoCohortError
from pydantic import Field


from ukrdc_stats.descriptions import dialysis_descriptions
from ukrdc_stats.models.generic_2d import (
    AxisLabels2d,
    Labelled2d,
    Labelled2dData,
    Labelled2dMetadata,
    BaseTable,
)

from ukrdc_stats.models.base import JSONModel
from ukrdc_stats.models.reports import CohortReport


class KRTMetadata(JSONModel):
    population: Optional[int] = Field(
        None,
        description="Number of patients in the cohort for dialysis stats calculation",
    )
    from_time: dt.datetime = Field(
        ..., description="Start time of dialysis stats calculations"
    )
    to_time: dt.datetime = Field(
        ..., description="End time of dialysis stats calculations"
    )


class KRTStats(JSONModel):
    """
    Container class for all the dialysis stats
    """

    incident_krt: Labelled2d = Field(
        ...,
        description="statistical breakdown of therapy types for incident patients in cohort",
    )
    prevalent_krt: Labelled2d = Field(
        ...,
        description="statistical breakdown of therapy types for prevalent patients in cohort",
    )
    incentre_dialysis_frequency: Labelled2d = Field(
        ...,
        description="per week frequency of dialysis for all in-centre dialysis patients",
    )
    incentre_time_dialysed: Labelled2d = Field(
        ...,
        description="per week time dialysed for all in-centre dialysis patients",
    )
    incident_initial_access: Labelled2d = Field(
        ...,
        description="vascular access of incident dialysis patients on their first session",
    )
    prevalent_most_recent_access: Labelled2d = Field(
        ...,
        description="vascular access of prevalent dialysis patients on their most recent session",
    )
    metadata: KRTMetadata


class UnitLevelKRTStats(JSONModel):
    all: KRTStats
    units: Dict[str, KRTStats]


def calculate_therapy_types(
    patient_cohort: pd.DataFrame,
) -> Tuple[List[str], List[int]]:
    """
    Breakdown of dialysis patients on home and in-centre therapies.
    The information is returned using pydantic classes designed to handle
    networks (this is essentially what a sankey plot is).

    Args:
        patient_cohort: DataFrame containing patient data.

    Returns:
        Tuple of two lists:
        - labels: A list of strings describing the type of therapy.
        - patients: A list of counts of patients for each type of therapy.
    """

    # Define mappings for 'qbl05' column
    mappings = {"HOSP": "In-centre", "SATL": "In-centre", "HOME": "Home"}

    # Update 'qbl05' based on conditions
    patient_cohort.loc[
        patient_cohort.registry_code_type.isin(["PD", "TX"]), "qbl05"
    ] = ""
    patient_cohort.loc[
        (patient_cohort.registry_code_type == "HD")
        & (patient_cohort.qbl05.isna() | (patient_cohort.qbl05 == "")),
        "qbl05",
    ] = "Unknown/Incomplete"
    patient_cohort.loc[:, "qbl05"] = patient_cohort["qbl05"].replace(mappings)

    # aggregate patients on modality and treatment supervision
    aggregate_patients = (
        patient_cohort.groupby(["registry_code_type", "qbl05"])["ukrdcid"]
        .count()
        .reset_index(name="count")
        .sort_values("registry_code_type")
    )

    # Create labels and patients lists
    labels = [
        f"{row.registry_code_type} {row.qbl05}".strip()
        for _, row in aggregate_patients.iterrows()
    ]
    patients = aggregate_patients["count"].tolist()

    return labels, patients


def adjust_next_fromtime(group: pd.DataFrame, **kwargs):
    """
    Utility function to adjust the next_fromtime in the case where there are
    overlaps in the treatment records.

    This function will blank the next_fromtime of all bar the last record in an
    overlapping group. Overlapping group ends when there is a gap not covered
    by treatment. The logic here is slightly fiddly.

    Args:
        group (_type_): _description_

    Returns:
        _type_: _description_
    """
    # skip any single record group
    group = group.reset_index(drop=True)
    if len(group) > 1:
        overlapping = False
        for i in range(len(group) - 1):
            if overlapping:
                ind_final = i
                to_time = group.at[i, "totime"]
                if to_time > max_to_time:  # noqa f841
                    max_to_time = group.at[i, "totime"]

                # overlap group ends where maximum to time in group is less
                # than the next from time.
                next_fromtime = group.at[i, "next_fromtime"]
                if max_to_time <= next_fromtime:
                    overlapping = False

                    # step 2: blank all next_fromtime in overlapping group
                    group.loc[ind_first:ind_final, "next_fromtime"] = pd.NaT  # noqa f821

                    # step 3: select record with maximum totime and add
                    # next_fromtime back in
                    overlap_slice = group[ind_first : ind_final + 1]  # noqa f821
                    max_totime_idx = overlap_slice["totime"].idxmax()
                    group.loc[max_totime_idx, "next_fromtime"] = next_fromtime

            else:
                # step 1: detect any overlapping records in group
                # locate the first record that satisfies the condition
                if group.at[i, "next_fromtime"] < group.at[i, "totime"]:
                    # create indices to track group
                    overlapping = True
                    ind_first = i  # noqa f841
                    ind_final = i
                    max_to_time = group.at[i, "totime"]

    return group


class KRTStatsCalculator(AbstractFacilityStatsCalculator):
    """Class to calculate basic statistics associated with the renal
    replacement therapies for renal facility in a given time window."""

    def __init__(
        self,
        session: Session,
        facility: str,
        date: dt.datetime = None,
        from_time: dt.datetime = None,
        to_time: dt.datetime = None,
    ):
        if to_time and date:
            date = to_time
        elif date:
            to_time = date
        else:
            date = to_time

        super().__init__(session, facility, date)

        if to_time > dt.datetime.now() - dt.timedelta(days=90):
            warnings.warn(
                "Stats calculated for times within the last 90 days may have their accuracy reduced",
            )

        self.recovery_window = 90  # days around the window to look at treatments

        if not from_time:
            from_time = to_time - dt.timedelta(days=365)

        # Create a precisely 2 element time window tuple
        self.time_window: Tuple[dt.datetime, dt.datetime] = (from_time, to_time)

        # defines encoding of KRT treatment types
        self.registry_code_types: List[str] = ["HD", "PD", "TX"]
        self.home_therapy_code_types: List[str] = ["HOSP", "SATL", "INCENTRE"]

    def _extract_patient_cohort(self, limit_to_ukrdc: Optional[bool] = True):
        """
        Extract a complete patient cohort dataframe. This is calculated fresh
        each time but we would probably want to implement some caching here.
        """
        self._patient_cohort = self._extract_incident_prevalent(
            self._extract_base_patient_cohort(limit_to_ukrdc=limit_to_ukrdc)
        )

    def _extract_base_patient_cohort(
        self,
        limit_to_ukrdc: Optional[bool] = True,
    ) -> pd.DataFrame:
        """Core query from which the other stats is derived. All patients at a
        renal facility with treatments up to and including 90 days post and
        prior to the time window will be included into the base cohort. Query
        will also flag any patients which had a historical ckd diagnosis or
        transplant. This should be rigorously back tested with real data and
        any changes should be considered breaking changes only to be done in a
        major release.

        Returns:
            pd.DataFrame: Patient cohort dataframe
        """

        minimum_transplant_length = 7

        ChronicTreatment = aliased(Treatment)  # pylint: disable=C0103
        ChronicModality = aliased(ModalityCodes)  # pylint: disable=C0103
        HistoricTransplantTreatment = aliased(Treatment)  # pylint: disable=C0103
        TransplantModality = aliased(ModalityCodes)  # pylint: disable=C0103
        SubPatientRecord = aliased(PatientRecord)

        # Select ukrdcids of patients treated at facility
        ukrdc_sub = select(PatientRecord.ukrdcid).where(
            PatientRecord.sendingfacility == self.facility
        )

        chronic_check = exists().where(
            ChronicTreatment.pid == SubPatientRecord.pid,
            SubPatientRecord.ukrdcid == PatientRecord.ukrdcid,
            ChronicTreatment.fromtime
            < self.time_window[1],  # Check if within time window
            ChronicTreatment.admitreasoncode
            == ChronicModality.registry_code,  # Match chronic modality code
            ChronicModality.registry_code_type == "CK",
        )

        tx_check = exists().where(
            HistoricTransplantTreatment.pid == SubPatientRecord.pid,
            SubPatientRecord.ukrdcid == PatientRecord.ukrdcid,
            HistoricTransplantTreatment.fromtime
            < self.time_window[0],  # Before start of time window
            HistoricTransplantTreatment.totime - HistoricTransplantTreatment.fromtime
            > dt.timedelta(days=minimum_transplant_length),  # Successful transplant
            HistoricTransplantTreatment.admitreasoncode
            == TransplantModality.registry_code,
            TransplantModality.registry_code_type == "TX",
        )

        if limit_to_ukrdc:
            ukrdc_sub.where(PatientRecord.sendingextract == "UKRDC")
            chronic_check.where(SubPatientRecord.sendingextract == "UKRDC")
            tx_check.where(SubPatientRecord.sendingextract == "UKRDC")

        query = (
            select(
                PatientRecord.pid,
                PatientRecord.ukrdcid,
                PatientRecord.sendingfacility,
                Treatment.healthcarefacilitycode,
                Treatment.admitreasoncode,
                Treatment.admitreasoncodestd,
                Treatment.admissionsourcecode,
                Treatment.admissionsourcecodestd,
                Treatment.qbl05,
                Treatment.hdp04,
                Treatment.dischargereasoncode,
                Treatment.dischargereasoncodestd,
                Treatment.dischargelocationcode,
                Treatment.dischargelocationcodestd,
                ModalityCodes.registry_code_type,
                Patient.deathtime,
                Treatment.fromtime,
                Treatment.totime,
                # Correlated subquery for chronic treatment check
                case(
                    (
                        chronic_check,
                        True,
                    ),
                    else_=False,
                ).label("is_chronic"),
                # Correlated subquery for historical transplant check
                case(
                    (
                        tx_check,
                        True,
                    ),
                    else_=False,
                ).label("historic_tx"),
            )
            .select_from(PatientRecord)
            .join(Patient, Patient.pid == PatientRecord.pid)
            .join(Treatment, Treatment.pid == PatientRecord.pid)
            .join(
                ModalityCodes, ModalityCodes.registry_code == Treatment.admitreasoncode
            )
            .where(
                ModalityCodes.registry_code_type.in_(self.registry_code_types),
                or_(
                    Treatment.totime > self.time_window[0] - dt.timedelta(days=90),
                    Treatment.totime.is_(None),
                ),
                or_(
                    Patient.deathtime > self.time_window[0], Patient.deathtime.is_(None)
                ),
                PatientRecord.ukrdcid.in_(ukrdc_sub),
            )
        )

        if dt.datetime.now() - self.time_window[1] > dt.timedelta(
            days=self.recovery_window
        ):
            query = query.where(
                Treatment.fromtime
                < self.time_window[1] + dt.timedelta(days=self.recovery_window)
            )

        if limit_to_ukrdc:
            query = query.where(PatientRecord.sendingextract == "UKRDC")

        # Execute query and explicitly specify the datatypes of columns
        base_cohort = pd.read_sql(
            query,
            self.session.bind,
            dtype={
                "pid": "string",
                "ukrdcid": "string",
                "sendingfacility": "string",
                "healthcarefacilitycode": "string",
                "admitreasoncode": "string",
                "admitreasoncodestd": "string",
                "admissionsourcecode": "string",
                "admissionsourcecodestd": "string",
                "qbl05": "string",
                "hdp04": "string",
                "dischargereasoncode": "string",
                "dischargereasoncodestd": "string",
                "dischargelocationcode": "string",
                "dischargelocationcodestd": "string",
                "registry_code_type": "string",
                "deathtime": "datetime64[ns]",
                "fromtime": "datetime64[ns]",
                "totime": "datetime64[ns]",
                "is_chronic": "bool",
                "historic_tx": "bool",
            },
        )

        if base_cohort.empty:
            raise NoCohortError(
                f"No patient cohort has been extracted. Facility {self.facility} may not have a UKRDC feed."
            )

        # pandas by default tries to be helpful and create compound keys
        # this is more overly helpful so we drop them
        base_cohort = base_cohort.reset_index(drop=True)

        return base_cohort

    def _chain_treatments(self, raw_patients: pd.DataFrame):
        """We append columns to the dataframe to allow recovery based
        calculations to be made.
        """

        raw_patients = raw_patients.sort_values(
            by=["ukrdcid", "fromtime", "sendingfacility"]
        )

        # append the start of the next treatment to each record
        raw_patients["next_fromtime"] = raw_patients.groupby("ukrdcid")[
            "fromtime"
        ].shift(-1)

        raw_patients = raw_patients.groupby("ukrdcid", as_index=False)[
            raw_patients.columns
        ].apply(adjust_next_fromtime, include_group=False)

        return raw_patients

    def _exclude_records(self, base_cohort: pd.DataFrame):
        """This implements any conditions the might cause a patient to be removed from
        the cohort. For example anyone who is in a 90 recovery period which spans the
        end of the time window should be excluded. Any patient with treatment modality
        code which implies CKD that dies before the end of the window will also be
        excluded.
        """

        # recovery window
        recoveries = (
            base_cohort["next_fromtime"] - base_cohort["totime"]
        ) > dt.timedelta(days=90)
        patient_recoveries = base_cohort[recoveries][
            ["ukrdcid", "next_fromtime", "totime"]
        ]

        index_to_remove = []
        for _, row in patient_recoveries.iterrows():
            if row["totime"] >= self.time_window[1]:
                # patient has made a recovery remove future records
                index_to_remove.extend(
                    base_cohort[
                        (base_cohort["ukrdcid"] == row["ukrdcid"])
                        & (base_cohort["fromtime"] > row["totime"])
                    ].index
                )
            else:
                if row["next_fromtime"] > self.time_window[1]:
                    # patient was recovered at end of window remove completely
                    index_to_remove.extend(
                        base_cohort[base_cohort["ukrdcid"] == row["ukrdcid"]].index
                    )

                else:
                    # patient coming out of recovery, remove record and all prior
                    index_to_remove.extend(
                        base_cohort[
                            (base_cohort["ukrdcid"] == row["ukrdcid"])
                            & (base_cohort["totime"] <= row["totime"])
                        ].index
                    )

        base_cohort = base_cohort.drop(index=index_to_remove)

        return base_cohort

    def _add_helper_columns(self, base_cohort: pd.DataFrame):
        """Function to postprocess data and add column to help with the
        calculation of incident and prevalent cohorts.

        Args:
            base_cohort (pd.DataFrame): Raw patient cohort generated by
            directly querying the database into pandas.

        Returns:
            _type_: _description_
        """

        # run function to link each treatment to the one that follows
        base_cohort = self._chain_treatments(base_cohort)

        # Exclude "acute" patients and records post or prior to recoveries
        base_cohort = self._exclude_records(base_cohort)

        # identify the records with the most recent from time for each ukrdcid
        most_recent = base_cohort[
            base_cohort["fromtime"] < self.time_window[1]
        ].reset_index(drop=True)
        most_recent = most_recent.groupby("ukrdcid", as_index=False)["fromtime"].max()

        # Merge with the original cohort to identify the most recent treatments
        base_cohort = base_cohort.merge(
            most_recent, on=["ukrdcid"], how="left", suffixes=("", "_max")
        )

        # Mark beginning and end of treatment timeline
        base_cohort["timeline_start"] = base_cohort.groupby("ukrdcid", as_index=False)[
            "fromtime"
        ].transform("min")
        base_cohort["timeline_stop"] = base_cohort.groupby("ukrdcid", as_index=False)[
            "totime"
        ].transform("max")

        # Mark first and most recent treatments
        base_cohort["first_treatment"] = (
            base_cohort["timeline_start"] == base_cohort["fromtime"]
        )
        base_cohort["most_recent"] = (
            base_cohort["timeline_stop"] == base_cohort["totime"]
        )

        # Calculate length of timeline
        # Null value here is >90days
        # TODO: this needs revisiting maybe a safer way of doing it is to
        # select whichever is larger now() - timeline_start or 91 days
        # this would be non deterministic
        base_cohort["timeline_length"] = (
            base_cohort["timeline_stop"] - base_cohort["timeline_start"]
        ).fillna(dt.timedelta(days=91))

        # Ensure deathtime exists for mocked/test data that may omit the column
        if "deathtime" not in base_cohort.columns:
            base_cohort["deathtime"] = pd.NaT

        base_cohort["life_length"] = (
            base_cohort["deathtime"] - base_cohort["timeline_start"]
        ).fillna(dt.timedelta(days=91))

        return base_cohort

    def _extract_incident_prevalent(self, base_cohort: pd.DataFrame) -> pd.DataFrame:
        """
        The function calculates the incident and prevalent cohorts as precisely
        as possible to the definition used in the annual report. However lack
        of full coverage means that transfer in patients will appear as
        incident patients. These will like lead to the incident cohorts to be
        an over estimate and prevalent cohorts to be an underestimate.

        Args:
            base_cohort (pd.DataFrame): Base cohort from output of _extract_base_patient_cohort
        Returns:
            pd.DataFrame: Patient cohort dataframe
        """

        # Generate some helper columns to label patients in helpful ways
        base_cohort = self._add_helper_columns(base_cohort)

        # Without full coverage we can do anything super accurate with transfer
        # out. However we will treat certain dischargereason codes as idicating
        # continued treatment.
        discharge_reasons = []  # = ["38"]?
        discharge_locations = ["ABROAD"]
        transfered_patients = base_cohort[
            (
                base_cohort["dischargelocationcode"].isin(discharge_locations)
                | base_cohort["dischargereasoncode"].isin(discharge_reasons)
            )
            & base_cohort.most_recent
        ].ukrdcid.drop_duplicates()
        transfered_out = base_cohort.ukrdcid.isin(transfered_patients)

        # Crash landed patients are defined:
        # - no chronic treatment records or tx
        # - remains on KRT for more than 90 days or transfered out
        # - survives for more than 90 days
        is_crash_landing = (
            (~base_cohort["is_chronic"] & ~base_cohort["historic_tx"])
            & (
                (base_cohort["timeline_length"] > dt.timedelta(days=90))
                | transfered_out
            )
            & (base_cohort["life_length"] > dt.timedelta(days=90))
        )

        # Patients with a previous record of transplant or ckd are considered
        # planned for KRT. These patients must stay on KRT for more than 90
        # days or die to be counted as incident.
        planned_ckd = (base_cohort["is_chronic"] | base_cohort["historic_tx"]) & (
            (base_cohort["timeline_length"] > dt.timedelta(days=90))
            | base_cohort["timeline_length"].isna()
            | transfered_out
        ) | (base_cohort["life_length"] < dt.timedelta(days=90))

        # Do we exclude patients which have had a historical transplant?
        # Also check treatments which turn up from different units.
        base_cohort["incident"] = (
            (planned_ckd | is_crash_landing)
            & (base_cohort["timeline_start"] > self.time_window[0])
            & (base_cohort["timeline_start"] <= self.time_window[1])
        )

        # Prevalence point defined at the end of the window patients are
        # counted as prevalent if their treatment timeline spans the end of the
        # window and is greater than 90 days.
        base_cohort["prevalent"] = (
            (base_cohort["timeline_start"] <= self.time_window[1])
            & (
                (base_cohort["timeline_stop"] > self.time_window[1])
                | base_cohort["timeline_stop"].isna()
            )
            & (base_cohort["timeline_length"] > dt.timedelta(days=90))
        )

        return base_cohort

    def _query_dialysis_sessions(
        self, patient_list: List[str], start: dt.datetime, stop: dt.datetime
    ) -> pd.DataFrame:
        # Calculate start of each week interval
        week_start = func.date_trunc("week", DialysisSession.procedure_time)

        dialysis_snomed = ["302497006", "233581009", "233586004"]

        query = (
            select(
                PatientRecord.pid,
                PatientRecord.ukrdcid,
                week_start.label("weekstart"),
                func.count(DialysisSession.procedure_type_code).label("hdsessionno"),
                func.sum(cast(DialysisSession.qhd31, Float)).label("totaltimedialised"),
            )
            .join(DialysisSession, DialysisSession.pid == PatientRecord.pid)
            .where(
                and_(
                    PatientRecord.pid.in_(patient_list),
                    DialysisSession.procedure_type_code.in_(
                        dialysis_snomed
                    ),  # filter for hd
                    DialysisSession.procedure_time > start,
                    DialysisSession.procedure_time < stop,
                )
            )
            .group_by(PatientRecord.pid, week_start)
        )

        session_data = pd.DataFrame(self.session.execute(query)).drop_duplicates()

        return session_data

    def _calculate_dialysis_frequency(self, subunit: str = "all") -> Labelled2d:
        """_summary_

        Args:
            subunit (str, optional): _description_. Defaults to "all".

        Returns:
            Labelled2d: _description_
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        patient_list = self._patient_cohort[
            (self._patient_cohort.registry_code_type == "HD")
            & (self._patient_cohort.qbl05.isin(["HOSP", "SATL", "In-centre"]))
            & self._patient_cohort.prevalent
        ]

        if subunit != "all":
            patient_list = patient_list[
                patient_list.healthcarefacilitycode == subunit
            ].pid.drop_duplicates()
        else:
            patient_list = patient_list.pid.drop_duplicates()

        dialysis_frequency_meta = Labelled2dMetadata(
            title="Median Haemodialysis Frequency",
            summary="Median frequency of incentre haemodialysis per week.",
            description=dialysis_descriptions["INCENTRE_DIALYSIS_FREQ"],
            axis_titles=AxisLabels2d(
                x="Frequency (days per week)", y="No. of Patients"
            ),
        )

        dialysis_time_meta = Labelled2dMetadata(
            title="Median Haemodialysis Time",
            summary="Median time of incentre haemodialysis per week.",
            description=dialysis_descriptions["INCENTRE_DIALYSIS_TIME"],
            axis_titles=AxisLabels2d(x="Time (hours per week)", y="No. of Patients"),
        )

        data_frequency = Labelled2dData(x=[], y=[])
        data_timedialised = Labelled2dData(x=[], y=[])
        if not patient_list.empty:
            # get number of dialysis sessions per patient and the date of the first and last one
            session_data = self._query_dialysis_sessions(
                patient_list, self.time_window[0], self.time_window[1]
            )
            if not session_data.empty:
                # drop any rows where hdsessionno is 0
                session_data = session_data[session_data["hdsessionno"] > 0]

                # Calculate the median of hdsessionno and timedialysed per ukrdcid
                median_data = (
                    session_data.groupby("ukrdcid")
                    .agg(
                        median_hdsessionno=pd.NamedAgg(
                            column="hdsessionno", aggfunc="median"
                        ),
                        median_timedialysed=pd.NamedAgg(
                            column="totaltimedialised", aggfunc="median"
                        ),
                    )
                    .reset_index()
                )

                freq_bins = [0.0, 1.0, 2.0, 3.0, 7.0]
                freq_labels = ["1", "2", "3", ">3"]

                frequency_hist = pd.cut(
                    median_data.median_hdsessionno, bins=freq_bins, labels=freq_labels
                ).value_counts(sort=False)

                time_bins = [0, 200, 400, 600, 800, 1000, 1000000]
                time_labels = [
                    "<200",
                    "201-400",
                    "401-600",
                    "601-800",
                    "801-1000",
                    ">1001",
                ]

                # create custom bins
                time_hist = pd.cut(
                    median_data.median_timedialysed, bins=time_bins, labels=time_labels
                ).value_counts(sort=False)

                # Update data_frequency with histogram data
                data_frequency.x = frequency_hist.index.tolist()
                data_frequency.y = frequency_hist.tolist()

                # Update data_timedialised with histogram data
                data_timedialised.x = time_hist.index.tolist()
                data_timedialised.y = time_hist.tolist()

        time_dialysis = Labelled2d(data=data_timedialised, metadata=dialysis_time_meta)
        frequency_dialysis = Labelled2d(
            data=data_frequency, metadata=dialysis_frequency_meta
        )

        return frequency_dialysis, time_dialysis

    def _calculate_median_dialysis_frequency(self, subunit: str = "all") -> Labelled2d:
        """Placeholder incase we revisit the idea of calculating the median
        dialysis frequency which would be more in line with what is calculated
        for the annual report.
        """
        del subunit
        return

    def _query_first_vascular_access(self, patient_list: pd.Series) -> pd.DataFrame:
        """
        SQL query to retreive the first vascular access used within a grace
        period of the first recorded dialysis session for a list of patients.

        Args:
            patient_list (pd.Series): List of pids defining a cohort.

        Returns:
            pd.DataFrame:  pid, accesstype, proceduretime
        """

        grace_period = 14
        time_windows = (
            select(
                DialysisSession.pid,
                DialysisSession.procedure_time.label("start_time"),
                (
                    DialysisSession.procedure_time + dt.timedelta(days=grace_period)
                ).label("end_time"),
            )
            .where(DialysisSession.pid.in_(patient_list))
            .distinct(DialysisSession.pid)
            .order_by(DialysisSession.pid, DialysisSession.procedure_time.asc())
        ).subquery()

        d = aliased(DialysisSession)
        vascular_access_query = (
            select(d.pid, d.qhd20, d.procedure_time)
            .join(
                time_windows,
                and_(
                    d.pid == time_windows.c.pid,
                    d.procedure_time.between(
                        time_windows.c.start_time, time_windows.c.end_time
                    ),
                ),
            )
            .where(d.qhd20.isnot(None))
            .distinct(d.pid)
            .order_by(d.pid, d.procedure_time.asc())
        )

        results = pd.DataFrame(self.session.execute(vascular_access_query))
        results = results.rename(columns={"qhd20": "accesstype"})

        return results

    def _query_most_recent_vascular_access(
        self, patient_list: pd.Series
    ) -> pd.DataFrame:
        """
        SQL query to retreive the most recent vascular access used within a grace
        period of the end date specified in the calculator.

        Args:
            patient_list (pd.Series): List of pids defining a cohort.

        Returns:
            pd.DataFrame: DataFrame with columns pid, accesstype, proceduretime
        """

        vascular_access_query = (
            select(
                DialysisSession.pid,
                DialysisSession.qhd20,
                DialysisSession.procedure_time,
            )
            .where(
                DialysisSession.pid.in_(patient_list),
                DialysisSession.procedure_time <= self.time_window[1],
                # DialysisSession.procedure_time >= self.time_window[1] - dt.timedelta(days=grace_period),
                DialysisSession.qhd20.isnot(None),
            )
            .distinct(DialysisSession.pid)
            .order_by(DialysisSession.pid, DialysisSession.procedure_time.desc())
        )

        results = pd.DataFrame(self.session.execute(vascular_access_query))
        results = results.rename(columns={"qhd20": "accesstype"})

        return results

    def _query_vascular_access(
        self, patient_list: pd.Series, is_first=True
    ) -> pd.DataFrame:
        """
        Function to return query either the first or the most recent vascular
        access for a list of patients. Patient list is batched to prevent the
        query size overflowing the maximum allowed.

        Args:
            patient_list (pd.Series): List of pids defining a cohort.
            is_first (bool, optional): Whether to return the first or most recent vascular access. Defaults to True.

        Returns:
            pd.DataFrame: DataFrame with columns pid, qhd20, proceduretime
        """
        CHUNK_SIZE = 100
        all_results = []

        pids = patient_list.tolist()

        for i in range(0, len(pids), CHUNK_SIZE):
            chunk_pids = pids[i : i + CHUNK_SIZE]

            if is_first:
                chunk_results = self._query_first_vascular_access(chunk_pids)
            else:
                chunk_results = self._query_most_recent_vascular_access(chunk_pids)

            if not chunk_results.empty:
                all_results.append(chunk_results)

        if all_results:
            vascular_access_data = pd.concat(all_results, ignore_index=True)
        else:
            vascular_access_data = pd.DataFrame(
                columns=["pid", "accesstype", "procedure_time"]
            )

        return vascular_access_data

    def _calculate_access_incident(self, subunit: str = "all") -> Labelled2d:
        """Displays the vascular access of incident patients on their first dialysis session
        Args:
            subunit (str, optional): Satellite unit. Defaults to "all".
        Raises:
            NoCohortError: e.g. if extract_patient_cohort has not been run
        Returns:
            Labelled2d: Number of incident patients with each type of access
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # filter by subunit
        if subunit == "all":
            patient_list = self._patient_cohort[
                self._patient_cohort.incident
            ].pid.drop_duplicates()
        else:
            patient_list = self._patient_cohort[
                self._patient_cohort.incident
                & (self._patient_cohort.healthcarefacilitycode == subunit)
                & self._patient_cohort.first_treatment
            ].pid.drop_duplicates()

        # function runs queries against the vascular access table and map to patients
        initial_access_data = self._query_vascular_access(patient_list)
        self._incident_access = self._patient_cohort[
            self._patient_cohort.incident
        ].merge(
            initial_access_data[["pid", "accesstype"]],
            on="pid",
            how="left",
        )

        self._incident_access.loc[
            self._incident_access.accesstype.isna(), "accesstype"
        ] = "Unknown/Incomplete"

        # aggregate cohort
        aggregate_patients = (
            self._incident_access.groupby(["accesstype"])["ukrdcid"]
            .count()
            .reset_index(name="value")
            .sort_values("accesstype")
        )
        x_data = list(aggregate_patients.accesstype)
        y_data = list(aggregate_patients.value)

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Vascular Access on First HD Session",
                summary="Vascular access for incident patients registered on their first dialysis session.",
                description=dialysis_descriptions["INCIDENT_INITIAL_ACCESS"],
                axis_titles=AxisLabels2d(x="Line Type", y="No. of Patients"),
                population_size=0,
            ),
            data=Labelled2dData(x=x_data, y=y_data),
        )

    def _calculate_access_prevalent(self, subunit: str = "all") -> Labelled2d:
        """Calculates the prevalent vascular access using the most recent access data.

        Args:
            subunit (str, optional): Satellite unit. Defaults to "all".

        Raises:
            NoCohortError: If no patient cohort has been extracted.

        Returns:
            Labelled2d: Number of prevalent patients with each type of access.
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # Filter by subunit
        if subunit == "all":
            patient_list = self._patient_cohort[
                self._patient_cohort.prevalent
            ].pid.drop_duplicates()
        else:
            patient_list = self._patient_cohort[
                self._patient_cohort.prevalent
                & (self._patient_cohort.healthcarefacilitycode == subunit)
            ].pid.drop_duplicates()

        # Query the most recent vascular access
        recent_access_data = self._query_vascular_access(patient_list, is_first=False)
        self._prevalent_access = self._patient_cohort[
            self._patient_cohort.prevalent
        ].merge(
            recent_access_data[["pid", "accesstype"]],
            on="pid",
            how="left",
        )

        self._prevalent_access.loc[
            self._prevalent_access.accesstype.isna(), "accesstype"
        ] = "Unknown/Incomplete"

        # Aggregate cohort
        aggregate_patients = (
            self._prevalent_access.groupby(["accesstype"])["ukrdcid"]
            .count()
            .reset_index(name="value")
            .sort_values("accesstype")
        )
        x_data = list(aggregate_patients.accesstype)
        y_data = list(aggregate_patients.value)

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Most Recent Vascular Access",
                summary="Vascular access for prevalent patients registered on their most recent dialysis session.",
                description=dialysis_descriptions["PREVALENT_MOST_RECENT_ACCESS"],
                axis_titles=AxisLabels2d(x="Line Type", y="No. of Patients"),
                population_size=0,
            ),
            data=Labelled2dData(x=x_data, y=y_data),
        )

    def _calculate_therapies_incident_patients(
        self, subunit: str = "all"
    ) -> Labelled2d:
        """Wrapper for calculate_therapy_types to calculate therapy types for an incident cohort
        Args:
            subunit (str, optional): Satellite unit. Defaults to "all".
        Raises:
            NoCohortError: _description_
        Returns:
            Labelled2d: Types of dialysis for incident patient cohort
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # TODO: add in something to assign patients to where they were first seen
        if subunit == "all":
            incident_cohort = self._patient_cohort[
                self._patient_cohort.incident & self._patient_cohort.first_treatment
            ]
        else:
            incident_cohort = self._patient_cohort[
                self._patient_cohort.incident
                & (self._patient_cohort.healthcarefacilitycode == subunit)
                & self._patient_cohort.first_treatment
            ]

        incident_labels, incident_no = calculate_therapy_types(incident_cohort)

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Incident KRT Modalities",
                summary="Breakdown of incident patients on PD and HD, and by home therapies and in-centre therapies.",
                description=dialysis_descriptions["INCIDENT_KRT_COHORT"],
                population_size=sum(incident_no),
            ),
            data=Labelled2dData(x=incident_labels, y=incident_no),
        )

    def _calculate_therapies_prevalent_patients(self, subunit: str = "all"):
        """Wrapper for calculate_therapy_types to calculate therapy types for an prevalent cohort
        Args:
            subunit (str, optional): Satellite unit. Defaults to "all".
        Raises:
            NoCohortError: _description_
        Returns:
            Labelled2d: Types of dialysis for prevalent patient cohort
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # filter patient cohort to get the first treatment type of each patient
        if subunit == "all":
            prevalent_cohort = self._patient_cohort[
                self._patient_cohort.prevalent & self._patient_cohort.first_treatment
            ]

        else:
            prevalent_cohort = self._patient_cohort[
                self._patient_cohort.prevalent
                # & self._patient_cohort.most_recent
                & (self._patient_cohort.healthcarefacilitycode == subunit)
            ]

        prevalent_labels, prevalent_no = calculate_therapy_types(prevalent_cohort)

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Prevalent KRT Modalities",
                summary="Breakdown of prevalent patients by PD and HD, and by home therapies and in-centre therapies.",
                description=dialysis_descriptions["PREVALENT_KRT_COHORT"],
                population_size=sum(prevalent_no),
            ),
            data=Labelled2dData(x=prevalent_labels, y=prevalent_no),
        )

    def extract_satellite_stats(self, unit: str = "all") -> KRTStats:
        """
        Returns:
            KRTStats:
        """

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # population size calculated from the sum of the incident and prevalant patients
        all_patients = self._patient_cohort[
            self._patient_cohort.incident | self._patient_cohort.prevalent
        ]
        if not unit == "all":
            all_patients = all_patients[all_patients.healthcarefacilitycode == unit]

        pop_size = len(all_patients.ukrdcid.unique())

        incident_krt = self._calculate_therapies_incident_patients(subunit=unit)
        prevalent_krt = self._calculate_therapies_prevalent_patients(subunit=unit)
        incentre_dialysis_frequency, incentre_time_dialysed = (
            self._calculate_dialysis_frequency(subunit=unit)
        )
        incident_initial_access = self._calculate_access_incident(subunit=unit)
        prevalent_most_recent_access = self._calculate_access_prevalent(subunit=unit)

        return KRTStats(
            metadata=KRTMetadata(
                population=pop_size,
                from_time=self.time_window[0],
                to_time=self.time_window[1],
            ),
            incident_krt=incident_krt,
            prevalent_krt=prevalent_krt,
            incentre_dialysis_frequency=incentre_dialysis_frequency,
            incentre_time_dialysed=incentre_time_dialysed,
            incident_initial_access=incident_initial_access,
            prevalent_most_recent_access=prevalent_most_recent_access,
        )

    def extract_stats(
        self,
        limit_to_ukrdc: Optional[bool] = True,
    ) -> UnitLevelKRTStats:
        """Extract all stats for the dialysis module
        Returns:
            KRTStats: Dialysis statistics object
        """
        # If we don't already have a patient cohort, extract one

        if self._patient_cohort is None:
            self.extract_patient_cohort(
                limit_to_ukrdc=limit_to_ukrdc,
            )

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        if self._patient_cohort.empty:
            raise NoCohortError(
                f"No patients found the cohort. Did you mean to try and extract facility {self.facility}?"
            )

        # calculate stats for all units
        unit_stats: Dict[str, KRTStats] = {}

        # loop over each unit and calculate stats
        for unit in _get_satellite_list(self.facility, self.session):
            unit_stats[unit] = self.extract_satellite_stats(unit)

        unit_stats[self.facility] = self.extract_satellite_stats(self.facility)
        return UnitLevelKRTStats(all=self.extract_satellite_stats(), units=unit_stats)

    def generate_cohort_report(
        self, cohort: str, include_ni: bool = False
    ) -> BaseTable:
        """
        Create a report to return the modalities of either incident or
        prevalent patients specified by the calculator.

        Args:
            cohort (str): "incident" or "prevalent"
            include_ni (bool, optional): Whether to include NI patients. Defaults to False.

        Returns:
            BaseTable: A table containing the modalities and incidence/prevalence
            status of patient for the cohort specified by the calculator.
        """

        # check the centre is in the output
        if cohort == "incident":
            pop, report = self.produce_report(
                [
                    "pid",
                    "healthcarefacilitycode",
                    "admitreasoncode",
                    "admitreasoncodestd",
                    "fromtime",
                    "totime",
                    "registry_code_type",
                ],
                [cohort, "first_treatment", f"sendingfacility == '{self.facility}'"],
                include_ni=include_ni,
            )
        elif cohort == "prevalent":
            # This needs reviewing since the proper definition of prevalence should be
            pop, report = self.produce_report(
                [
                    "pid",
                    "healthcarefacilitycode",
                    "admitreasoncode",
                    "admitreasoncodestd",
                    "fromtime",
                    "totime",
                    "registry_code_type",
                ],
                [cohort, "first_treatment", f"sendingfacility == '{self.facility}'"],
                include_ni=include_ni,
            )

        desc = f"Report on the treatment modalities of the {cohort} cohort. This report contains a table which includes the most recent treatment modality and the way it's classified by the renal registry along with the ukrdc patient identifier."

        return CohortReport(
            description=desc, cohort=cohort, population=pop, table=report
        )



"""
Functions to calculate patient cohort statistics.
Pydantic classes to return nice plotly friendly output.
Utility functions to handle bitty things.
"""

from ukrdc_stats.calculators.demographics import DemographicStatsCalculator
from ukrdc_stats.calculators.krt import KRTStatsCalculator

__all__ = ["DemographicStatsCalculator", "KRTStatsCalculator"]



"""Calculators associated with ckd. In particular the care planning."""

from operator import and_
import pandas as pd
import datetime as dt
from sqlalchemy import select, or_, create_engine, tuple_, case, func
from sqlalchemy.orm import Session, sessionmaker, aliased

from ukrdc_sqla.ukrdc import (
    Patient,
    PatientRecord,
    Treatment,
    Address,
    PatientNumber,
    ResultItem,
    LabOrder,
    CodeMap,
    ModalityCodes,
)
from ukrdc_sqla.xmlarchive import (
    Patient as XMLPatient,
    Assessment,
    Treatment as XMLTreatment,
)

from ukrdc_stats.calculators.abc import AbstractFacilityStatsCalculator
from ukrdc_stats.models.generic_2d import BaseTable
from ukrdc_stats.models.base import JSONModel
from ukrdc_stats.utils import egfr


def get_archive_session(session: Session) -> Session:
    # function to return a session of the xml v5 archive db
    # (xmlschemaconverter storage) not sure why this doesn't work
    db_url = session.bind.url

    password = db_url.password
    username = db_url.username
    host = db_url.host
    port = db_url.port
    drivername = db_url.drivername
    database = "removed_xml_archive"

    new_url = f"{drivername}://{username}:{password}@{host}:{port}/{database}"
    engine = create_engine(new_url)
    session = sessionmaker(bind=engine)

    with session() as archive_session:
        return archive_session


class CarePlanningReport(JSONModel):
    description: str = "prevalent ckd cohort..."
    cohort: str = "Prevalent CKD"
    table: BaseTable


class PrevalentCKDCalculator(AbstractFacilityStatsCalculator):
    def __init__(
        self,
        session: Session,
        facility: str,
        prevalence_point: dt.datetime = dt.datetime.now(),
        v5_archive_session: Session = None,
    ):
        super().__init__(session, facility)
        self._prevalence_point = prevalence_point

        if not v5_archive_session:
            self.v5_archive_session = get_archive_session(session)
        else:
            self.v5_archive_session = v5_archive_session

        self._ckd_not_rrt_codes = ["901", "902", "903"]

    def extract_stats(self):
        pass

    def _core_query(self, extract_all: bool = False):
        # get a single address per patient with a preference for home address
        address_ranked = (
            select(
                Address.pid,
                Address.postcode,
                Address.addressuse,
                func.row_number()
                .over(
                    partition_by=Address.pid,
                    order_by=case((Address.addressuse == "H", 0), else_=1),
                )
                .label("rn"),
            ).where(Address.postcode.is_not(None), func.trim(Address.postcode) != "")
        ).subquery()

        # Create an alias for Treatment to join on itslef later
        Treatment2 = aliased(Treatment)

        query_ckd_patients = (
            select(
                PatientRecord.pid,
                PatientRecord.ukrdcid,
                PatientRecord.sendingfacility,
                Patient.birthtime,
                Patient.deathtime,
                Treatment.healthcarefacilitycode,
                Treatment.healthcarefacilitydesc,
                Treatment.admitreasoncode,
                Treatment.admitreasoncodestd,
                Treatment.admitreasondesc,
                Treatment.fromtime,
                Treatment.totime,
                Patient.gender.label("sex"),
                address_ranked.c.postcode,
                address_ranked.c.addressuse,
                Patient.ethnicgroupcode,
                Patient.ethnicgroupdesc,
                CodeMap.destination_code.label("ukkaethnicity"),
                ModalityCodes.registry_code_type,
            )
            .join(Treatment, Treatment.pid == PatientRecord.pid)
            .join(Patient, Patient.pid == PatientRecord.pid)
            .outerjoin(address_ranked, address_ranked.c.pid == PatientRecord.pid)
            .outerjoin(
                CodeMap,
                and_(
                    CodeMap.source_code == Patient.ethnicgroupcode,
                    CodeMap.source_coding_standard == Patient.ethnicgroupcodestd,
                ),
            )
            .join(PatientNumber, PatientNumber.pid == PatientRecord.pid, isouter=True)
            .where(
                ModalityCodes.registry_code_type.in_(["CK", "CN"]),
                PatientRecord.sendingfacility == self.facility,
                PatientRecord.sendingextract == "UKRDC",
                or_(
                    Patient.deathtime > self._prevalence_point,
                    Patient.deathtime.is_(None),
                ),
                or_(
                    CodeMap.destination_coding_standard == "URTS_ETHNIC_GROUPING",
                    CodeMap.destination_coding_standard.is_(None),
                ),
                address_ranked.c.rn == 1,
            )
        )

        if extract_all:
            query_ckd_patients = (
                query_ckd_patients.join(Treatment2, Treatment2.pid == Treatment.pid)
                .join(
                    ModalityCodes,
                    ModalityCodes.registry_code == Treatment2.admitreasoncode,
                )
                .where(
                    Treatment2.fromtime < self._prevalence_point,
                    or_(
                        Treatment2.totime > self._prevalence_point,
                        Treatment2.totime.is_(None),
                    ),
                )
            )
        else:
            query_ckd_patients = query_ckd_patients.join(
                ModalityCodes, ModalityCodes.registry_code == Treatment.admitreasoncode
            ).where(
                Treatment.fromtime < self._prevalence_point,
                or_(
                    Treatment.totime > self._prevalence_point,
                    Treatment.totime.is_(None),
                ),
            )

        query_ckd_patients = query_ckd_patients.order_by(PatientRecord.pid)

        base_cohort = (
            pd.DataFrame(self.session.execute(query_ckd_patients))
            .drop_duplicates()
            .reset_index(drop=True)
        )

        if not extract_all:
            base_cohort = base_cohort.sort_values(
                ["pid", "fromtime"], ascending=[True, False]
            ).drop_duplicates("pid", keep="first")

        return base_cohort

    def _get_patient_numbers(self, pids: list[str]) -> pd.DataFrame:
        CHUNK_SIZE = 100
        all_patient_numbers = []

        # Process PIDs in specified chunks
        for i in range(0, len(pids), CHUNK_SIZE):
            chunk_pids = pids[i : i + CHUNK_SIZE]

            query = select(
                PatientNumber.pid,
                PatientNumber.patientid,
                PatientNumber.organization,
                PatientNumber.numbertype,
            ).where(
                PatientNumber.pid.in_(chunk_pids),
            )

            chunk_results = pd.DataFrame(self.session.execute(query)).drop_duplicates()
            if not chunk_results.empty:
                all_patient_numbers.append(chunk_results)

        # Combine all chunks
        if all_patient_numbers:
            patients_numbers = pd.concat(all_patient_numbers, ignore_index=True)
        else:
            patients_numbers = pd.DataFrame(
                columns=["pid", "patientid", "organization", "numbertype"]
            )

        return patients_numbers.reset_index(drop=True).astype(str)

    def _get_archive_data(
        self, patient_numbers: pd.DataFrame, extract_all: bool = False
    ):
        # Break up large queries into chunks to avoid PostgreSQL stack overflow
        BATCH_SIZE = 100
        all_assessments = []
        all_treatments = []

        # Process patient numbers in batches
        for i in range(0, len(patient_numbers), BATCH_SIZE):
            batch = patient_numbers.iloc[i : i + BATCH_SIZE]

            # Assessments query for this batch
            assessments_query = (
                select(
                    XMLPatient.nationalid.label("patientid"),
                    XMLPatient.organization,
                    XMLPatient.numbertype,
                    XMLPatient.creation_date,
                    Assessment.assessmentstart,
                    Assessment.assessmentend,
                    Assessment.assessmenttypecode,
                    Assessment.assessmenttypecodestd,
                    Assessment.assessmenttypecodedesc,
                    Assessment.assessmentoutcomecode,
                    Assessment.assessmentoutcomecodestd,
                    Assessment.assessmentoutcomecodedesc,
                )
                .join(
                    Assessment,
                    Assessment.patientid == XMLPatient.id,
                )
                .where(
                    Assessment.assessmentstart < self._prevalence_point,
                    tuple_(
                        XMLPatient.nationalid,
                        XMLPatient.organization,
                        XMLPatient.numbertype,
                    ).in_(
                        list(
                            zip(
                                batch["patientid"],
                                batch["organization"],
                                batch["numbertype"],
                            )
                        )
                    ),
                )
            )

            # Treatments query for this batch
            treatments_query = (
                select(
                    XMLPatient.nationalid.label("patientid"),
                    XMLPatient.organization,
                    XMLPatient.numbertype,
                    XMLPatient.creation_date,
                    XMLTreatment.admitreasoncode,
                    XMLTreatment.admitreasoncodestd,
                    XMLTreatment.admitreasondesc,
                    XMLTreatment.fromtime,
                    XMLTreatment.totime,
                )
                .join(
                    XMLTreatment,
                    XMLTreatment.patientid == XMLPatient.id,
                )
                .where(
                    XMLTreatment.admitreasoncode.in_(self._ckd_not_rrt_codes),
                    tuple_(
                        XMLPatient.nationalid,
                        XMLPatient.organization,
                        XMLPatient.numbertype,
                    ).in_(
                        list(
                            zip(
                                batch["patientid"],
                                batch["organization"],
                                batch["numbertype"],
                            )
                        )
                    ),
                )
            )

            if not extract_all:
                treatments_query = treatments_query.where(
                    XMLTreatment.fromtime < self._prevalence_point,
                    or_(
                        XMLTreatment.totime > self._prevalence_point,
                        XMLTreatment.totime.is_(None),
                    ),
                )

            # Execute queries and collect results
            batch_assessments = pd.DataFrame(
                self.v5_archive_session.execute(assessments_query)
            )
            if not batch_assessments.empty:
                all_assessments.append(batch_assessments)

            batch_treatments = pd.DataFrame(
                self.v5_archive_session.execute(treatments_query)
            )
            if not batch_treatments.empty:
                all_treatments.append(batch_treatments)

        # Combine results from all batches
        if all_assessments:
            assessments = pd.concat(all_assessments).reset_index(drop=True)
        else:
            assessments = pd.DataFrame(
                columns=[
                    "patientid",
                    "organization",
                    "numbertype",
                    "creation_date",
                    "assessmentstart",
                    "assessmentend",
                    "assessmenttypecode",
                    "assessmenttypecodestd",
                    "assessmenttypecodedesc",
                    "assessmentoutcomecode",
                    "assessmentoutcomecodestd",
                    "assessmentoutcomecodedesc",
                ]
            )

        if all_treatments:
            treatments = pd.concat(all_treatments).reset_index(drop=True)
        else:
            treatments = pd.DataFrame(
                columns=[
                    "patientid",
                    "organization",
                    "numbertype",
                    "creation_date",
                    "admitreasoncode",
                    "admitreasoncodestd",
                    "admitreasondesc",
                    "fromtime",
                    "totime",
                ]
            )

        # drop ids and deduplicate (incase same patient has been written multiple times)
        assessments = pd.merge(
            assessments,
            patient_numbers,
            on=["patientid", "organization", "numbertype"],
            how="inner",
        )
        treatments = pd.merge(
            treatments,
            patient_numbers,
            on=["patientid", "organization", "numbertype"],
            how="inner",
        )

        assessments = assessments.drop(
            columns=["patientid", "organization", "numbertype"]
        ).drop_duplicates()
        treatments = treatments.drop(
            columns=["patientid", "organization", "numbertype"]
        ).drop_duplicates()

        return treatments, assessments

    def _get_test_results(self, patient_ids):
        """gets the most recent creatinine and lab egfr
        "QBLA1", "QBLAB", "QBLAL"
        Args:
            patient_ids (_type_):
        """

        query = (
            select(
                LabOrder.pid,
                ResultItem.serviceidcode,
                ResultItem.resultvalue,
                ResultItem.resultvalueunits,
                ResultItem.observationtime,
            )
            .distinct(LabOrder.pid, ResultItem.serviceidcode)
            .join(LabOrder, LabOrder.id == ResultItem.order_id)
            .where(
                ResultItem.serviceidcode.in_(["QBLA1", "QBLAB", "QBLAL"]),
                LabOrder.pid.in_(patient_ids),
                ResultItem.observation_time < self._prevalence_point,
            )
            .order_by(
                LabOrder.pid,
                ResultItem.serviceidcode,
                ResultItem.observation_time.desc(),
            )
        )

        results = pd.DataFrame(self.session.execute(query).all())
        if results.empty:
            columns = [
                "pid",
                "serviceidcode",
                "resultvalue",
                "resultvalueunits",
                "observationtime",
            ]
            results = pd.DataFrame(columns=columns)

        # separate and clean
        egfr_results = results[results["serviceidcode"].isin(["QBLAB", "QBLAL"])].copy()
        egfr_results["resultvalue"] = (
            egfr_results["resultvalue"].str.replace("<", "").str.replace(">", "")
        )
        egfr_results["resultvalue"] = pd.to_numeric(
            egfr_results["resultvalue"], errors="coerce"
        )

        # Drop rows where eGFR conversion to numeric failed then deduplicate
        egfr_results = egfr_results.dropna(subset=["resultvalue"])

        egfr_results = egfr_results.sort_values("observationtime").drop_duplicates(
            subset=["pid"], keep="last"
        )

        # Get creatinine results
        creatinine_results = results[
            results["serviceidcode"] == "QBLA1"
        ]  # .drop(columns=['serviceidcode'])
        creatinine_results["resultvalue"] = pd.to_numeric(
            creatinine_results["resultvalue"], errors="coerce"
        )

        # Merge creatinine and eGFR results
        merged_results = pd.merge(
            creatinine_results,
            egfr_results,
            on="pid",
            how="outer",
            suffixes=("_creat", "_labegfr"),
        )

        return merged_results

    def _extract_base_patient_cohort(self, extract_all: bool = False):
        # Get main cohort from ukrdc
        cohort = self._core_query(extract_all)

        if cohort.empty:
            return

        # Get all know patient identifiers for matching
        patient_numbers = self._get_patient_numbers(cohort["pid"].tolist())

        # Send patient numbers to the archive to extract data from there
        treatments, assessments = self._get_archive_data(patient_numbers, extract_all)

        # correct the treatments using the archive data
        # we assume treatments without corresponding ukrdc record are invalid
        # this could/should be restricted to codes that can map to eachother
        # e.g. 902 -> 900 where like '9%' or something
        cohort = pd.merge(
            cohort,
            treatments,
            on=["pid", "fromtime", "totime"],
            how="left",
            suffixes=("_ukrdc", ""),
        )

        # add in treatments not in the archive and drop ukrdc values
        ukrdc_only = cohort["admitreasoncode"].isnull()
        cohort.loc[
            ukrdc_only, ["admitreasoncode", "admitreasoncodestd", "admitreasondesc"]
        ] = cohort.loc[
            ukrdc_only,
            [
                "admitreasoncode_ukrdc",
                "admitreasoncodestd_ukrdc",
                "admitreasondesc_ukrdc",
            ],
        ].values
        cohort = cohort.drop(
            columns=[
                "admitreasoncode_ukrdc",
                "admitreasoncodestd_ukrdc",
                "admitreasondesc_ukrdc",
            ]
        )

        # join assessments
        cohort = pd.merge(
            cohort,
            assessments,
            on=["pid"],
            how="left",
            suffixes=("_treatment", "_assessment"),
        )

        test_results = self._get_test_results(cohort["pid"].tolist())

        # get test results and cakculate egfr
        cohort = pd.merge(
            cohort,
            test_results,
            on=["pid"],
            how="left",
        )

        cohort["calculated_egfr"] = cohort.apply(
            lambda row: egfr(
                row["resultvalue_creat"],
                row["resultvalueunits_creat"],
                row["observationtime_creat"],
                row["birthtime"],
                row["sex"],
                row["ukkaethnicity"],
            ),
            axis=1,
        )

        # Get universal patient ids (NHS number ideally)
        patient_ids_sorted = (
            patient_numbers.drop(columns=["numbertype"])
            .sort_values(
                by=["organization"],
                key=lambda x: x.map({"NHS": 0, "CHI": 1, "HSC": 2, "LOCALHOSP": 3}),
            )
            .drop_duplicates(subset=["pid"], keep="first")
        )

        # Merge patient them back into the cohort
        cohort = pd.merge(
            cohort,
            patient_ids_sorted,
            on=["pid"],
            how="left",
        )

        cohort.rename(columns={"patientid": "externalid"}, inplace=True)

        return cohort

    def _extract_patient_cohort(self, extract_all: bool = True):
        self._patient_cohort = self._extract_base_patient_cohort(extract_all)



"""
Common utility functions useful in multiple statistics
"""

import datetime as dt
from ukrdc_sqla.ukrdc import Code
import pandas as pd
import fileinput
import warnings

from ukrdc_sqla.ukrdc import CodeMap, SatelliteMap
from sqlalchemy.orm import Session
from sqlalchemy import select, and_
from typing import Optional, Dict, List


def egfr(
    scr: float,
    scr_unit: str,
    scr_date: dt.datetime,
    dob: dt.datetime,
    sex: int = 1,
    ethnicity: Optional[str] = None,
) -> Optional[int]:
    """Function for calculating the egfr based on the equation found here:
    http://nephron.com/epi_equation

    Args:
        scr (int): serum creatinine level
        scr_unit (str): unit of serum creatinine
        scr_date (dt.datetime): date of serum creatinine measurement
        dob (dt.datetime): date of birth
        sex (int, optional): sex of patient. Defaults to 1 (male).
        ethnicity (Optional[str], optional): ethnicity of patient. Defaults to None.

    Returns:
        Optional[int]: estimated glomerular filtration rate
    """

    if pd.isna(scr) or pd.isna(scr_date):
        return

    age = age_from_dob_exact(scr_date, dob)

    if age < 18:
        return

    # only accept creatinines with accepted units
    if scr_unit == "umol/L":
        scr = scr / 88.4
    elif scr_unit == "mmol/L":
        scr = scr / (10 * 88.4)
    elif scr_unit == "g/L":
        scr = 100.0 * scr
    elif scr_unit == "mg/dL":
        pass
    else:
        return

    if sex == "2":
        kappa = 0.7
        alpha = -0.329
        multiplier = 1.018
    else:
        kappa = 0.9
        alpha = -0.411
        multiplier = 1.0

    scr_frac = scr / kappa
    if scr_frac > 1:
        multiplier = multiplier * (scr_frac**-1.209)
    else:
        multiplier = multiplier * (scr_frac**alpha)

    egfr = round(141 * multiplier * (0.993**age))

    return egfr


def age_from_dob(date: dt.date, dob: dt.date) -> int:
    """Returns the age on a given date

    Args:
        date (datetime): Date to calculate age or time period from.
        dob (datetime): Date to calculate age or time period at.

    Returns:
        int: age or period in years
    """
    years_old: int

    # calculates age by common definition
    years_old = date.year - dob.year
    if (dob.month == 2) & (dob.day == 29):
        # handles case where birthday is on leap day
        year_birthday = dt.datetime(date.year, dob.month, dob.day - 1)
    else:
        year_birthday = dt.datetime(date.year, dob.month, dob.day)

    if year_birthday > date:
        years_old -= 1

    return years_old


def age_from_dob_exact(date: dt.date, dob: dt.date) -> float:
    """Generates an exact dob as decimal

    Args:
        date (dt.date): Date to calculate age or time period from.
        dob (dt.date): Date to calculate age or time period at.

    Returns:
        float: age
    """

    return (date - dob).days / 365.25


def dob_cutoff_from_age(date: dt.datetime, age: int) -> dt.datetime:
    """returns a date a fixed number of years before give date

    Args:
        date (dt.date): date to calculate from
        age (int): number of years before date

    Returns:
        dt.date: date a set number of years ago
    """

    return date - dt.timedelta(days=age * 365.25)


def map_codes(source_std: str, destination_std: str, session: Session) -> dict:
    """Use the code map table to return a code mapping set from the ukrdc as a
    dictionary.

    Args:
        source_std (str): _description_
        destination_std (str): _description_
        session (Session): _description_

    Returns:
        dict: _description_
    """

    query = select(CodeMap.source_code, CodeMap.destination_code).where(
        and_(
            CodeMap.source_coding_standard == source_std,
            CodeMap.destination_coding_standard == destination_std,
        )
    )

    codes = pd.DataFrame(session.execute(query))

    return dict(zip(codes.source_code, codes.destination_code))


def lookup_codes(
    coding_standard: str, attribute: str, session: Session
) -> Dict[str, str]:
    """Get a code set from the ukrdc lookup and return some attribute from it
    (most likely the description)

    Args:
        coding_standard (str): The coding standard to lookup
        attribute (str): The attribute to return (e.g., 'description')
        session (Session): SQLAlchemy database session

    Returns:
        Dict[str, str]: Dictionary mapping code values to the requested attribute
    """
    # Build and execute query properly
    query = select(Code).where(Code.coding_standard == coding_standard)
    result = session.execute(query).scalars().all()

    # Handle empty results
    if not result:
        warnings.warn(f"No codes found for coding standard '{coding_standard}'")
        return {}

    # Convert to dictionary directly from ORM objects
    return {code.code: getattr(code, attribute, None) for code in result}


def strip_whitespace(filepath: str):
    """Run to stop pylint complaining about trailing whitespace"""

    for line in fileinput.input(filepath, inplace=True):
        line = line.rstrip()
        if line:
            print(line)


def _mapped_key(key: str) -> str:
    """Tiny convenience function to return a common mapped column name

    Args:
        key (str): Column to map

    Returns:
        str: Mapped column name
    """
    return f"{key}_mapped"


def _calculate_base_patient_histogram(
    cohort: pd.DataFrame, group: str, code_map: Optional[Dict[str, str]] = None
) -> pd.DataFrame:
    """Extract a histogram of the patient cohort, grouped by the given column

    Args:
        cohort (pd.DataFrame): Patient cohort
        group (str): Column to group by

    Raises:
        NoCohortError: If the patient cohort is empty

    Returns:
        pd.DataFrame: Histogram dataframe of the patient cohort
    """

    if code_map:
        mapped_column = _mapped_key(group)
        cohort[mapped_column] = cohort[group].map(code_map)

        histogram = (
            cohort[["ukrdcid", mapped_column]]
            .drop_duplicates()
            .groupby([mapped_column])
            .count()
            .reset_index()
        )

    else:
        histogram = (
            cohort[["ukrdcid", group]]
            .drop_duplicates()
            .groupby([group])
            .count()
            .reset_index()
        )

    return histogram.rename(columns={"ukrdcid": "Count"})


def _mapped_if_exists(df: pd.DataFrame, column: str) -> pd.Series:
    """
    Convenience function to return the mapped column if it exists,
    otherwise return the original column

    Args:
        df (pd.DataFrame): Input dataframe
        column (str): Column to return

    Returns:
        pd.Series: Mapped column if it exists, otherwise the original column
    """
    mapped_column: str = _mapped_key(column)
    if mapped_column in df.columns:
        return df[mapped_column]
    else:
        warnings.warn(
            f"Column {mapped_column} does not exist in dataframe, returning {column} instead"
        )
        return df[column]


def _get_satellite_list(facility_code: str, session: Session) -> List[str]:
    """
    Get the list of satellites for the facility.
    """
    query = select(SatelliteMap.satellite_code).where(
        SatelliteMap.main_unit_code == facility_code
    )
    return session.execute(query).scalars().all()


def check_headcounts(cohort: pd.DataFrame, groupby_attributes: list[str] = []):
    """Used in the scripts to ensure headcounts remain consistent and patients
    aren't being dropped

    Args:
        cohort (pd.DataFrame): dataframe with coumns group_byattributes + value
        groupby_attributes (_type_, optional): columns to label data e.g centre

    Raises:
        Warning: _description_
    """

    if not groupby_attributes:
        groupby_attributes = ["satellite_code", "centre_code", "variable2"]

    # remove paeds to keep things simple
    if "adultpaed" in cohort.columns:
        cohort = cohort[cohort["adultpaed"] == "Adult"]

    # aggregate over specified columns
    head_count = (
        cohort.groupby(groupby_attributes)
        .sum(numeric_only=True)
        .reset_index()[groupby_attributes + ["value"]]
    )

    # intialise some bits for use later
    label_columns = groupby_attributes.copy()
    if "variable2" in label_columns:
        label_columns.remove("variable2")

    previous_labels = None
    previous_value = None
    msg = None

    # cross check rows for consistency
    for _, row in head_count.iterrows():
        labels = row[label_columns].to_list()
        value = row["value"]

        if labels == previous_labels:
            if value != previous_value:
                msg = "Headcount id variable across categories\n"
                mask = (
                    head_count[label_columns] == pd.Series(labels, index=label_columns)
                ).all(axis=1)
                msg += head_count[mask].to_string(index=False) + "\n"
        else:
            previous_labels = labels
            previous_value = value

    if msg:
        raise Warning(msg)

    return


def aggregate_data(
    dataframe: pd.DataFrame, groupby_attributes: list[str]
) -> pd.DataFrame:
    """Simple utility function to wrap pandas aggregation on ukrdcid

    Args:
        dataframe (pd.DataFrame): _description_
        groupby_attributes (list[str]): _description_
        value_column (str): _description_
    """

    return (
        dataframe.drop_duplicates(subset=["ukrdcid"])
        .groupby(groupby_attributes)
        .size()
        .reset_index(name="value")
    )


VASCULAR_MAPPING = {
    "AVF": "AVF/AVG",
    "AVFUO": "AVF/AVG",
    "AVG": "AVF/AVG",
    "TLN": "TL",
    "NLN": "NTL",
    "HER": "AVF/AVG",
}

# NHS digital gender map
GENDER_GROUP_MAP = {"1": "Male", "2": "Female", "9": "Indeterminate", "X": "Unknown"}



from ukrdc_stats.models.base import JSONModel
from ukrdc_stats.models.generic_2d import BaseTable


class CohortReport(JSONModel):
    description: str
    cohort: str
    population: int
    table: BaseTable



"""
Module to contain the long descriptions for the pydantic output
"""

from textwrap import dedent


dialysis_general = """
- Patients are queried on the basis of their treatment records overlapping the time window. 
- 
"""

dialysis_descriptions = {
    "ALL_PATIENTS_KRT_COHORT": dedent(
        """
        # All Patients Undergoing Kidney Replacement Therapy

        ## Overview
        This pie chart illustrates the proportion of patients who received kidney replacement therapy within the time period. The chart is broken down by the type of treatment, including HD In-center, HD Home, HD Unknown/Incomplete, PD, and Tx. Optionally the chart can be filtered by satellite unit. 

        ## Treatment Definitions
        - HD: Haemodialysis patients (with a modality defined as HD by the UKRDC). This includes patients registered for haemodialysis, haemofiltration, haemodiafiltration, or ultrafiltration. 
        - PD: Peritoneal dialysis (with a modality defined as PD by the UKRDC).This includes patients registered for CAPD or APD treatments.
        - TX: Transplant patients (with a modality defined as TX), including both living and cadaver donors.
        - In-centre: HD patients with qbl05 field of the Treatment table as HOSP or SATL.   
        - Home: HD patients with qbl05 field of the Treatment table as HOME. 
        - Unknown/Incomplete: HD patients with incomplete qbl05 field or anything other than HOME, HOSP, or SATL

        ## Methodology 
        - Any patients with a time of death before the beginning of the time window were excluded from the cohort, as were any patients whose treatments started before and ended after it.
        - Patient's therapy types was selected using the admission reason and the unit, and were further split into home and in-center therapy groups (with all patients on PD included in the home therapies group).
        - The numbers were calculated from aggregating patients within the five groups: HD Home, HD In-centre, HD Unknown/Incomplete, PD and TX.
        - No deduplication is applied to the treatment records so patients with multiple treatments will be double counted
                
        ## UKRDC Entities Used
        The chart was produced by joining the following UKRDC entities according to their foreign key relationships:
        - [PatientRecord](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450149/PatientRecord): ukrdcid, sendingextract
        - [Patient](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450145/Patient): deathtime
        - [Treatment](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450155/Treatment+Encounter): qbl05, hdp04, fromtime, totime, dischargereasoncode, healthcarefacilitycode
        - [ModalityCodes](https://renalregistry.atlassian.net/l/cp/Ac1YeFfH): registry_code_type
        """
    ),
    "INCIDENT_KRT_COHORT": dedent(
        """
        # Incident Kidney Replacement Patients

        ## Definition
        A patient starting kidney replacement therapy (KRT) - defined as haemodialysis, peritoneal dialysis, 
        or kidney transplant - for the first time during the selected time period.

        ## Inclusion Criteria
        1. First treatment starts within the selected dates. This is defined as the first treatment preceded by a gap of greater than 90 days without KRT treatment.
        2. Either:
           - Known kidney disease history (planned start)
           - No prior history (unplanned/"crash" start) AND survives >90 days
        3. Treatment continues for at least 90 days OR patient:
           - Has planned start and dies within 90 days
           - Transfers to another unit
        4. Patient is modality is assigned to the first 

        ## Timeline Example
        ```
        Key:
        X = Treatment
        - = No Treatment
        * = Death
        T = Transfer

        90 days prior                                                     90 days after
        |<--------------------->|<----------------------->|<----------------->|
        [Gap Check]             [Start]                   [End]           [Follow-up] 
            
        Incident          ------|---XXXXXXXXXXXXXXXXXXXXXX|XXXXX  (Continuous after start)
        Incident          XXX---|---------XXXXXXXXXXXXXXXX|XXXXX  (>90 day gap timeline resets)
        Incident          CKD---|--XXXX*------------------|-----  (Dies within 90 days and prior history)
        Incident          ------|---XXX--XXX-XXXXXXXXX-XXX|XXXXX  (Discontinuous with short gaps)
        Incident          ------|-----------------------XX|XXXXX  (Begins at end of window and continues)
        Incident          ------|---XXXXXXXXXXXXXXXXT-----|-----  (Treatment ends with transfer out)
        
        Not Incident 
        Not Incident      ------|--XXXX*------------------|-----  (Dies within 90 days and no prior history)
        Not incident      XXX---|-XXXXXXXXXXXXXXXXXXXXXXXX|XXXXX  (<90 day gap and timeline begins prior)
        ```

        ## UKRDC Entities Used
        - [PatientRecord](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450149/PatientRecord): ukrdcid, sendingextract
        - [Patient](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450145/Patient): deathtime
        - [Treatment](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450155/Treatment+Encounter): qbl05, hdp04, fromtime, totime, dischargereasoncode, healthcarefacilitycode
        - [ModalityCodes](https://renalregistry.atlassian.net/l/cp/Ac1YeFfH): registry_code_type

        ## Data Quality & Limitations
        The software used to calculate the statistics should be considered experimental and is subject to the following non-exhaustive limitations:
        1. Recent Window Effects (<90 days follow-up):
           - Cannot confirm treatment continuation
           - May count some patients receiving acute treatment

        2. Data Coverage:
           - Inter-unit transfers appear as new starts
           - These may inflate incidence rates

        3. Data Quality:
           - UKRDC contains uncleaned data and may be incomplete or contain errors
        """
    ),
    "PREVALENT_KRT_COHORT": dedent(
        """
        # Prevalent Kidney Replacement Patients

        ## Definition
        A patient receiving ongoing kidney replacement therapy (KRT) - defined as haemodialysis, peritoneal dialysis, 
        or kidney transplant - who is established on treatment at the end of the selected time period.

        ## Inclusion Criteria
        1. Treatment continues beyond end of selected dates
        2. Has received at least 90 days of treatment before window end
        3. No gaps in treatment greater than 90 days
        4. Either:
            - Active treatment at window end
            - Recent transfer to another unit (discharge code 38)

        ## Timeline Example 
        ```
        Example Cases:
        X = Treatment
        - = No Treatment
        * = Death
        T = Transfer
        
                                                        Prevalence Point                                        
        |<-------------------->|<--------------------------->|
                                       [Analysis Window]      

        Prevalent        XXXXXX|XXXXXXXXXXXXXXXXXXXXXXXXXXXXX|X----  (Active at end)
        Prevalent        XXXXXX|XXXXXXXXXXXXXXXXXXXXXXXXT--->|>----  (Transfer out)
        Prevalent        --XXXX|XXXXXXXX------------XXXXXXXXX|X----  (>90 days at end)
        Prevalent        ------|----------------------------X|XXXXX  (>90 days at end)
        
        Not Prevalent    XXXXXX|XXXXXXXXXXXXXXX*-------------|-----  (Died in window)
        Not Prevalent    ------|----------------------XXXX---|-X---  (<90 days total)
        ```    

        ## UKRDC Entities Used
        The chart was produced by joining the following UKRDC entities according to their foreign key relationships:
        - [PatientRecord](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450149/PatientRecord): ukrdcid, sendingextract
        - [Patient](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450145/Patient): deathtime
        - [Treatment](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2006450155/Treatment+Encounter): qbl05, hdp04, fromtime, totime, dischargereasoncode, healthcarefacilitycode
        - [ModalityCodes](https://renalregistry.atlassian.net/l/cp/Ac1YeFfH): registry_code_type

        ## Data Quality & Limitations
        The software used to calculate the statistics should be considered experimental and is subject to the following non-exhaustive limitations:
        
        1. Data Coverage:
        - Patients from another unit may be counted as prevalent if they are being treated in the unit at prevalence point

        2. Data Quality:
        - UKRDC contains uncleaned data and may be incomplete or contain errors

    """
    ),
    "INCENTRE_DIALYSIS_FREQ": dedent(
        """
        # In-Centre Dialysis Frequency

        ## Overview
        This histogram represents the mean number of dialysis sessions per week for all dialysis patients in a three month period at a sendingfacility or one of its satellites. Optionally the chart can be filtered by satellite unit. 

        ## Methodology
        - Dialysis sessions are counted for patients in the 'All Patients Undergoing Kidney Replacement Therapy' cohort. This is done by grouping on the procedure type code. 
        - Patients with less than two sessions are rejected. 
        - The per week frequency is calculated for each person by dividing the count by the time difference between their first and last dialysis session within the three month period.
        - Patients are aggregated into bins of with boundaries (0.0, 0.5, 1.5, 2.5, 3.5, 7.0). This are labelled 1,2,3 and >3 sessions per week.  

        ## UKRDC Entities Used
        The dialysis sessions table is queried by grouping by ukrdcid with the following aggregate functions used:
        - https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2005565449/Dialysis+Session+Procedure: MIN(fromtime), MAX(totime), COUNT(sessiontype).
        """
    ),
    "INCENTRE_DIALYSIS_TIME": dedent(
        """
        # In-Centre Dialysis Time

        ## Overview
        """
    ),
    "INCIDENT_INITIAL_ACCESS": dedent(
        """
        # Incident Initial Access
        ## Overview
        This pie chart shows the vascular access recorded on the first dialysis session of each incident patient. Optionally the chart can be filtered by satellite unit.

        ## Methodology
        - This cohort is identical to that used for incident patients treatment breakdown.
        - The first session is selected by ranking the dialysis sessions in time.
        - If the type of vascular access has been recorded it is counted accordingly otherwise it is counted as Unknown/Incomplete.
       
       ## UKRDC Entities Used
        - [Dialysis Session](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2005565449/Dialysis+Session+Procedure) qhd20 for first dialysis session. 
        """
    ),
    "PREVALENT_MOST_RECENT_ACCESS": dedent(
        """
        # Prevalent Most Recent Access
        ## Overview
        Vascular access recorded on the most recent dialysis session of each prevalent patient.

        ## Methodology
        - This cohort is identical to that used for prevalent patients treatment breakdown.
        - The most recent coded vascular access is selected within a two week period of the end of the calculation window
        - ipsum lopsum nausium
       
       ## UKRDC Entities Used
        - [Dialysis Session](https://renalregistry.atlassian.net/wiki/spaces/UD/pages/2005565449/Dialysis+Session+Procedure) qhd20 for first dialysis session. 
        """
    ),
}

demographic_generic = """
        - The initial cohort is selected by filtering the patientrecords sent via the ukrdc.  
        - The cohort is filtered to only include patients with a treatment record which overlaps 90 days prior to the calculation time. 
"""

demographic_descriptions = {
    "GENDER_DESCRIPTION": dedent(
        f"""
        # Patient Gender
        Gender identity recorded for each living patient registered with the renal unit. Specifically it is defined as the person stated gender code in the nhs data dictionary
        
        # Methodology
        {demographic_generic}
        - Patient records are matched to NHS stated gender using patient demographic information 
        - Patients are optionally checked against NHS tracing to check for date of death
        - All living patients with patient records sent by a particular sending facility are aggregated based on gender


        ## UKRDC Entities Used
        - [PatientRecord](https://renalregistry.atlassian.net/l/cp/KCZ6A2bX)
        - [Patient](https://renalregistry.atlassian.net/l/cp/0MXHtpTU)
        
        """
    ),
    "ETHNIC_GROUP_DESCRIPTION": dedent(
        f"""
        # Patient Ethnicity

        ## Overview 
        Ethnicity group code recorded for each living patient registered with the renal unit over all time.
        The five ethnicity groupings used to map ethnicity codes onto the displayed ethnicity values are the same as those used in the Renal Registry Annual Report.
        
        ## Methodology
        {demographic_generic}
        - Patient records are matched to ethnicity using patient demographic information 
        - Patients are optionally checked against NHS tracing to check for date of death
        - All living patients with patient records sent by a particular sending facility are aggregated based on ethnicity

        ## UKRDC Entities Used
        - [PatientRecord](https://renalregistry.atlassian.net/l/cp/KCZ6A2bX)
        - [Patient](https://renalregistry.atlassian.net/l/cp/0MXHtpTU)
        
        """
    ),
    "AGE_DESCRIPTION": dedent(
        f"""
        # Patient Age
        The age, calculated from date of birth, recorded for each living patient registered with the renal unit.
        
        # Methodology
        {demographic_generic}
        - Patient records are matched to date of birth using patient demographic information 
        - Age is calculated from date of birth 
        - Patients are optionally checked against NHS tracing to check for date of death
        - All living patients with patient records sent by a particular sending facility are aggregated based on age
        
        ## UKRDC Entities Used
        - [PatientRecord](https://renalregistry.atlassian.net/l/cp/KCZ6A2bX)
        - [Patient](https://renalregistry.atlassian.net/l/cp/0MXHtpTU)
        """
    ),
}



"""
Generic 2D data model for UKRDC stats
"""

import csv

import pandas as pd

from datetime import datetime
from typing import List, Optional, Any
from pydantic import Field

from ukrdc_stats.models.base import JSONModel

Number = int | float | None
RowData = List[Any]


# Generics
class AxisLabels2d(JSONModel):
    """
    Generic class for any x/y axis labels
    """

    x: Optional[str] = None
    y: Optional[str] = None


class Basic2dMetadata(JSONModel):
    """
    Stats/plot metadata generic to all 2D data sets
    """

    title: str = Field(..., description="Title of plot or statistics", max_length=40)
    summary: str = Field(
        ...,
        description="Summary of what the plot or statistic is trying to achieve",
        max_length=100,
    )
    description: str = Field(
        ...,
        description="In depth description of what exactly is represented and how it has been calculated",
    )
    axis_titles: Optional[AxisLabels2d] = Field(
        None, description="x and y labels for data"
    )
    population_size: Optional[int] = Field(
        None,
        description="Total population size of cohort used for statistic e.g total number of patients in pie chart",
    )


# Time series


class TimeSeries2dData(JSONModel):
    """
    x-y data for a time series plot
    """

    x: List[datetime] = Field(..., description="datatime labels for datapoints")
    y: List[Number] = Field(..., description="Datapoints")
    error_y: Optional[List[Number]] = Field(
        None, description="error bars associated with datapoints"
    )


class TimeSeries2dMetadata(Basic2dMetadata):
    """
    Metadata for a time series plot.

    x-axis is always a datetime, so no x-axis units are required.
    """

    units_y: Optional[str] = Field(None, description="units of y datapoints")


class TimeSeries2d(JSONModel):
    """
    Return data class for a time series plot
    """

    metadata: TimeSeries2dMetadata = Field(
        ...,
        description="Metadata including title and description for 2D time series dataset",
    )
    data: TimeSeries2dData = Field(
        ..., description="2D dataset of timestamped datapoints"
    )


# Numeric


class Numeric2dData(JSONModel):
    """
    x-y data for a numeric plot
    """

    x: List[Number] = Field(..., description="Numeric x data points")
    y: List[Number] = Field(..., description="Y data points")
    error_x: Optional[List[Number]] = Field(
        None, description="Uncertainty associated with x data points"
    )
    error_y: Optional[List[Number]] = Field(
        None, description="Uncertainty associated with y data points"
    )


class Numeric2dMetadata(Basic2dMetadata):
    """
    Metadata for a numeric plot.

    We may want units for both x and y values, so both are optional here.
    """

    units_x: Optional[str] = Field(None, description="Units of y datapoints")
    units_y: Optional[str] = Field(None, description="Units of x datapoints")


class Numeric2d(JSONModel):
    """
    Return data class for a numeric plot
    """

    metadata: Numeric2dMetadata = Field(
        ...,
        description="Metadata including title and description for 2D time series dataset",
    )
    data: Numeric2dData = Field(
        ..., description="2D data containing x and y as numbers"
    )


class Labelled2dData(JSONModel):
    """
    x-y data for a labelled plot
    """

    x: List[str] = Field(..., description="list of x data points")
    y: List[Number] = Field(..., description="list of y data points")
    error_y: Optional[List[Number]] = Field(
        None, description="Uncertainty in y data points"
    )


class Labelled2dMetadata(Basic2dMetadata):
    """
    Metadata for a labelled plot.

    x-axis is always a string/label, so no x-axis units are required.
    y-axis is numeric, and so units are optional here.
    """

    coding_standard_x: Optional[str] = Field(
        None, description="UKRDC coding standard of x data points"
    )  # I wonder if this should be a list otherwise are assuming coding standard remains constant
    units_y: Optional[str] = Field(None, description="Units of y data point")


class Labelled2d(JSONModel):
    """
    Return data class for a labelled plot
    """

    metadata: Labelled2dMetadata = Field(
        ..., description="Metadata for 2D data consisting of label datapoint pairs"
    )
    data: Labelled2dData = Field(
        ..., description="2D data consisting of label datapoint pairs"
    )


class BaseTable(JSONModel):
    headers: List[str] = Field(
        ..., description="Column headers describing the data contained in each row"
    )

    rows: List[RowData] = Field(
        ..., description="Rows of the table containing the data"
    )
    # rows: List[str] = Field(
    #    ..., description="Rows of the table containing the data"
    # )

    def to_csv(
        self, file_path: str, blank_na: bool = True, metadata: str = None
    ) -> None:
        """
        Serializes the BaseTable to a CSV file.

        :param file_path: The path to the file where the CSV data will be written.
        """
        with open(file_path, mode="w", newline="", encoding="utf-8") as csv_file:
            if metadata:
                csv_file.write(metadata)
            writer = csv.writer(csv_file)
            writer.writerow(self.headers)  # Write the headers
            for row in self.rows:
                writer.writerow(
                    [item if not pd.isna(item) else "" for item in row]
                    if blank_na
                    else row
                )

    def to_pandas(self) -> None:
        return pd.DataFrame(self.rows, columns=self.headers).astype("string")
